/**
 * Copyright (C) 2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dmenu;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.dussan.vaadin.dmenu.client.rpc.MenuServerRpc;
import org.dussan.vaadin.dmenu.client.state.MenuState;
import org.dussan.vaadin.dmenu.helper.ManifestHelper;
import org.dussan.vaadin.dmenu.menuitem.MenuItem;
import org.dussan.vaadin.dmenu.menuitem.MenuItemElement;
import org.dussan.vaadin.dtabs.DTabs;
import org.dussan.vaadin.dtabs.DTabsProperty;

import com.vaadin.server.Resource;
import com.vaadin.ui.Component;
import com.vaadin.ui.HorizontalLayout;

public class DMenu
    extends DTabs
{

    private static final long serialVersionUID = -1766784516812336397L;

    private static final String VERSION = "Implementation-Version";

    private static final String GIT_VERSION = "Git-Version";

    private static final String MENU_ITEM_ID = "dmenu-";

    private static final String MENU_ITEM_IS_EMPTY = "Menu item is empty.";

    private static final String MENU_ALREADY_HAVE_THIS_ITEM = "Menu already have this item.";

    private static final String INDEX_HAVE_TO_BE_GREATER_THAN_OR_EQUAL_0 = "Index have to be greater than or equal 0.";

    private static final String COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM = "Component is not instance of MenuItem.";

    private boolean menuItemsSameHeight = false;

    private boolean menuItemElementsSameHeight = false;

    private boolean menuTabsAutoChange = false;

    private boolean floatingMenu = false;

    private boolean floatingMenuItemAutoHide = false;

    private boolean floatingMenuItemAutoShow = false;

    private int menuItemCounter = 0;

    private int selectedMenuItem = 0;

    private List<MenuItem> menuItems = null;

    public DMenu()
    {
        menuItems = new ArrayList<>();
        setProperties( DTabsProperty.COMPACTED_TABBAR, DTabsProperty.FRAMED_TABS );
        setMenuItemsSameHeight( true );
        setMenuItemElementsSameHeight( true );

        registerRpc( new MenuServerRpc()
        {
            private static final long serialVersionUID = 9039890628414457601L;

            @Override
            public void menuActiveTabChange( String activeTabId )
            {
                for ( int index = 0; index < getMenuItems().size(); index++ )
                {
                    if ( activeTabId.equals( getTab( index ).getStyleName() ) )
                    {
                        setSelectedMenuItem( index );
                        break;
                    }
                }
            }

            @Override
            public void menuItemShow( String menuItemId )
            {
                for ( int index = 0; index < getMenuItems().size(); index++ )
                {
                    if ( menuItemId.equals( getTab( index ).getStyleName() ) )
                    {
                        setSelectedMenuItem( index );
                        break;
                    }
                }
            }
        } );

        addSelectedTabChangeListener( new SelectedTabChangeListener()
        {
            private static final long serialVersionUID = -1047609548802091735L;

            @Override
            public void selectedTabChange( SelectedTabChangeEvent event )
            {
                setSelectedMenuItem( getMenuItemIndex( event.getTabSheet()
                                                            .getSelectedTab() ) );
            }
        } );
    }

    public DMenu( MenuItem... menuItems )
    {
        this();
        addMenuItems( menuItems );
    }

    public int getSelectedMenuItem()
    {
        selectedMenuItem = getMenuItemIndex( getSelectedTab() );
        return selectedMenuItem;
    }

    public void setSelectedMenuItem( int position )
    {
        setSelectedTab( Math.max( 0, Math.min( position, menuItems.size() - 1 ) ) );
    }

    public void setSelectedMenuItem( MenuItem menuItem )
    {
        setSelectedTab( getMenuItemIndex( menuItem ) );
    }

    public boolean isMenuItemsSameHeight()
    {
        return menuItemsSameHeight;
    }

    public void setMenuItemsSameHeight( boolean menuItemsSameHeight )
    {
        if ( isMenuItemsSameHeight() != menuItemsSameHeight )
        {
            this.menuItemsSameHeight = menuItemsSameHeight;
            getState( true ).setMenuItemsSameHeight( menuItemsSameHeight );
        }
    }

    public boolean isMenuItemElementsSameHeight()
    {
        return menuItemElementsSameHeight;
    }

    public void setMenuItemElementsSameHeight( boolean menuItemElementsSameHeight )
    {
        if ( isMenuItemElementsSameHeight() != menuItemElementsSameHeight )
        {
            this.menuItemElementsSameHeight = menuItemElementsSameHeight;
            getState( true ).setMenuItemElementsSameHeight( menuItemElementsSameHeight );
        }
    }

    public boolean isMenuTabsAutoChange()
    {
        return menuTabsAutoChange;
    }

    public void setMenuTabsAutoChange( boolean menuTabsAutoShow )
    {
        if ( isMenuTabsAutoChange() != menuTabsAutoShow )
        {
            this.menuTabsAutoChange = menuTabsAutoShow;
            getState( true ).setMenuItemAutoChange( menuTabsAutoShow );
        }
    }

    public boolean isFloatingMenu()
    {
        return floatingMenu;
    }

    public void setFloatingMenu( boolean floatingMenu )
    {
        if ( isFloatingMenu() != floatingMenu )
        {
            this.floatingMenu = floatingMenu;
            getState( true ).setFloatingMenu( floatingMenu );
        }
    }

    public boolean isFloatingMenuItemAutoHide()
    {
        return floatingMenuItemAutoHide;
    }

    public void setFloatingMenuItemAutoHide( boolean floatingMenuItemAutoHide )
    {
        if ( isFloatingMenuItemAutoHide() != floatingMenuItemAutoHide )
        {
            this.floatingMenuItemAutoHide = floatingMenuItemAutoHide;
            getState( true ).setFloatingMenuItemAutoHide( floatingMenuItemAutoHide );
        }
    }

    public boolean isFloatingMenuItemAutoShow()
    {
        return floatingMenuItemAutoShow;
    }

    public void setFloatingMenuItemAutoShow( boolean floatingMenuItemAutoShow )
    {
        if ( isFloatingMenuItemAutoShow() != floatingMenuItemAutoShow )
        {
            this.floatingMenuItemAutoShow = floatingMenuItemAutoShow;
            getState( true ).setFloatingMenuItemAutoShow( floatingMenuItemAutoShow );
        }
    }

    public List<MenuItem> getMenuItems()
    {
        return menuItems;
    }

    public void addMenuItems( MenuItem... menuItems )
    {
        for ( MenuItem menuItem : menuItems )
        {
            addMenuItem( menuItem );
        }
    }

    public int getMenuItemIndex( Component component )
    {
        return menuItems.indexOf( component );
    }

    public MenuItem getMenuItem( int index )
    {
        return menuItems.get( index );
    }

    public void addMenuItem( MenuItem menuItem )
    {
        if ( menuItem == null || menuItem.isEmpty() )
        {
            throw new IllegalArgumentException( MENU_ITEM_IS_EMPTY );
        }
        else if ( menuItems.contains( menuItem ) )
        {
            throw new IllegalArgumentException( MENU_ALREADY_HAVE_THIS_ITEM );
        }

        addMenuItem( menuItem, null, menuItems.size() );
    }

    public void addMenuItem( MenuItem menuItem, int position )
    {
        if ( menuItem == null || menuItem.isEmpty() )
        {
            throw new IllegalArgumentException( MENU_ITEM_IS_EMPTY );
        }
        else if ( menuItems.contains( menuItem ) )
        {
            throw new IllegalArgumentException( MENU_ALREADY_HAVE_THIS_ITEM );
        }
        else if ( position < 0 )
        {
            throw new IllegalArgumentException( INDEX_HAVE_TO_BE_GREATER_THAN_OR_EQUAL_0 );
        }

        addMenuItem( menuItem, null, position );
    }

    public void addMenuItem( MenuItem menuItem, Resource icon, int position )
    {
        if ( menuItem == null || menuItem.isEmpty() )
        {
            throw new IllegalArgumentException( MENU_ITEM_IS_EMPTY );
        }
        else if ( menuItems.contains( menuItem ) )
        {
            throw new IllegalArgumentException( MENU_ALREADY_HAVE_THIS_ITEM );
        }
        else if ( position < 0 )
        {
            throw new IllegalArgumentException( INDEX_HAVE_TO_BE_GREATER_THAN_OR_EQUAL_0 );
        }

        menuItem.setParentMenuItemIndex( position );
        menuItems.add( position, menuItem );
        super.addTab( menuItem, menuItem.getCaption(), icon, position );
        super.getTab( menuItem ).setStyleName( MENU_ITEM_ID + menuItemCounter++ );
    }

    public void removeMenuItem( int position )
    {
        menuItems.remove( position );
        super.removeTab( super.getTab( position ) );
    }

    public boolean isEmpty()
    {
        return menuItems.isEmpty();
    }

    public static String getVersion()
        throws IOException
    {
        if ( ManifestHelper.getManifest() != null )
        {
            return ManifestHelper.getManifest()
                                 .getMainAttributes()
                                 .getValue( VERSION );
        }
        return null;
    }

    public static String getGitVersion()
        throws IOException
    {
        if ( ManifestHelper.getManifest() != null )
        {
            return ManifestHelper.getManifest()
                                 .getMainAttributes()
                                 .getValue( GIT_VERSION );
        }
        return null;
    }

    @Override
    protected MenuState getState()
    {
        return (MenuState) super.getState();
    }

    @Override
    protected MenuState getState( boolean markAsDirty )
    {
        return (MenuState) super.getState( markAsDirty );
    }

    @Override
    public Tab addTab( Component c )
    {
        if ( !( c instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        return super.addTab( c );
    }

    @Override
    public Tab addTab( Component c, String caption )
    {
        if ( !( c instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        return super.addTab( c, caption );
    }

    @Override
    public Tab addTab( Component c, String caption, Resource icon )
    {
        if ( !( c instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        return super.addTab( c, caption, icon );
    }

    @Override
    public Tab addTab( Component component, int position )
    {
        if ( !( component instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        return super.addTab( component, position );
    }

    @Override
    public Tab addTab( Component tabComponent, String caption, Resource icon, int position )
    {
        if ( !( tabComponent instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        return super.addTab( tabComponent, caption, icon, position );
    }

    @Override
    public void addComponent( Component c )
    {
        if ( !( c instanceof MenuItem ) )
        {
            throw new IllegalArgumentException( COMPONENT_IS_NOT_INSTANCE_OF_MENU_ITEM );
        }
        addMenuItem( (MenuItem) c );
    }

    @Override
    public void addComponents( Component... components )
    {
        for ( Component component : components )
        {
            addComponent( component );
        }
    }

    @Override
    public void beforeClientResponse( boolean initial )
    {
        super.beforeClientResponse( initial );
        getState( true ).setSelectedMenuItem( getSelectedMenuItem() );

        if ( isMenuItemsSameHeight() )
        {
            getState( true ).setMenuItemsSameHeightDefined( !initial );
            if ( initial )
            {
                getState( true ).setMenuItemElementsCount( menuItems.get( getSelectedMenuItem() )
                                                                    .getMenuItemElements()
                                                                    .size() );

                // move all menu item elements to active menu item
                // so we can get max height of menu item elements
                HorizontalLayout layout = (HorizontalLayout) super.getTab( getSelectedMenuItem() ).getComponent();
                for ( int index = 0; index < getComponentCount(); index++ )
                {
                    if ( index != getSelectedMenuItem() )
                    {
                        layout.moveComponentsFrom( (HorizontalLayout) super.getTab( index ).getComponent() );
                    }
                }
            }
            else
            {
                // move menu item elements to menu item which it belongs
                for ( int index = 0; index < menuItems.size(); index++ )
                {
                    for ( MenuItemElement menuItemElement : getMenuItem( index ).getMenuItemElements() )
                    {
                        if ( menuItemElement.getParentMenuItemIndex() != index )
                        {
                            menuItems.get( menuItemElement.getParentMenuItemIndex() )
                                     .addComponent( menuItemElement );
                        }
                    }
                }

                getState( true ).setMenuItemElementsCount( menuItems.get( getSelectedMenuItem() )
                                                                    .getMenuItemElements()
                                                                    .size() );
            }
        }
    }

}
